import json
import logging

import httpx
from httpx import Response

from libqcanvas_clients.canvas.canvas_client_config import CanvasClientConfig
from libqcanvas_clients.util.authenticator import Authenticator


class CanvasAuthenticator(Authenticator):
    _logger = logging.getLogger(__name__)

    def __init__(self, client: httpx.AsyncClient, client_config: CanvasClientConfig):
        super().__init__(client)
        self._canvas_client_config = client_config

    async def _authenticate(self) -> None:
        self._logger.info("Attempting authentication to canvas")

        legacy_authentication_url = await self.request_legacy_authentication_url()
        response = await self._client.get(legacy_authentication_url)

        if not response.is_redirect:
            self._logger.error("Authentication failed %s", response)
            raise RuntimeError("Authentication failed")
        else:
            self._logger.info("Authentication was successful")

    async def request_legacy_authentication_url(self) -> str:
        response = await self._client.get(
            url=self._canvas_client_config.get_endpoint("login/session_token"),
            headers=self._canvas_client_config.get_authorization_header(),
        )

        if response.is_success:
            return self._extract_legacy_authentication_url_from_response(response)
        else:
            self._logger.warning("Your API key may be incorrect")
            raise RuntimeError("Authentication failed, check your API key")

    def _extract_legacy_authentication_url_from_response(
        self, session_response: Response
    ) -> str:
        target_url = json.loads(session_response.text)["session_url"]

        self._logger.debug("Authenticating via %s", target_url)

        if target_url is None:
            self._logger.warning(
                "Session response body was malformed. Got: %s", session_response.text
            )
            raise ValueError("Session response body was malformed")
        else:
            return target_url



import logging
from typing import Any, Optional, override

import gql
import httpx
from gql.client import DocumentNode as GraphQLQuery
from graphql import ExecutionResult
from httpx import URL, Cookies, Request, Response
from tenacity import (
    retry,
    retry_if_exception_type,
    stop_after_attempt,
    wait_exponential_jitter,
)

from libqcanvas_clients.canvas.canvas_authenticator import CanvasAuthenticator
from libqcanvas_clients.canvas.canvas_client_config import CanvasClientConfig
from libqcanvas_clients.canvas.canvas_types import (
    DiscussionTopicHeader,
    LegacyPage,
    MediaObject,
    RemoteFile,
)
from libqcanvas_clients.canvas.canvas_unauthenticated_detector_hook import (
    CanvasUnauthenticatedDetectorHook,
)
from libqcanvas_clients.util.add_cookies_to_request_hook import AddCookiesToRequestHook
from libqcanvas_clients.util.external_client_httpx_transport import (
    ExternalClientHTTPXTransport,
)
from libqcanvas_clients.util.json_util import check_status_and_parse_json
from libqcanvas_clients.util.request_exceptions import (
    ConfigInvalidError,
    RatelimitedError,
)


class CanvasClient:
    _logger = logging.getLogger(__name__)

    def __init__(
        self,
        client_config: CanvasClientConfig,
        max_concurrent_operations: int = 20,
        gql_timeout: float = 120,
        httpx_timeout: float = 120,
        client: Any = None,
    ):
        self._client_config = client_config
        self._client = client or httpx.AsyncClient(
            timeout=httpx_timeout,
            limits=httpx.Limits(max_connections=max_concurrent_operations),
            http2=True,
        )
        self._authentication_controller = CanvasAuthenticator(
            self._client, self._client_config
        )
        self._gql_timeout = gql_timeout
        self._gql_client = gql.Client(
            transport=_CanvasTransport(
                client=self._client,
                url=self._client_config.get_endpoint("api/graphql"),
                headers=self._client_config.get_authorization_header(),
            ),
            execute_timeout=self._gql_timeout,
        )

        self._add_cookie_request_hook(self._client)
        self._add_authentication_detector_hook(self._client)

    def _add_cookie_request_hook(self, client: httpx.AsyncClient) -> None:
        client.event_hooks["request"].append(
            AddCookiesToRequestHook(self._client.cookies.jar)
        )

    def _add_authentication_detector_hook(self, client: httpx.AsyncClient) -> None:
        client.event_hooks["response"].append(
            CanvasUnauthenticatedDetectorHook(self._client_config.host)
        )

    @retry(
        wait=wait_exponential_jitter(initial=0.5, exp_base=1.2, max=10),
        retry=retry_if_exception_type(RatelimitedError),
        stop=stop_after_attempt(8),
    )
    async def graphql_query(
        self,
        query: GraphQLQuery | str,
        query_variables: Optional[dict[str, Any]] = None,
    ) -> dict[str, Any]:
        if query_variables is None:
            query_variables = {}

        if not isinstance(query, GraphQLQuery):
            query = gql.gql(query)

        return await self._gql_client.execute_async(
            document=query, variable_values=query_variables
        )

    async def get_temporary_session_url(self) -> str:
        return await self._authentication_controller.request_legacy_authentication_url()

    async def get_current_user_id(self) -> str:
        self._logger.debug("Fetching current user id")
        response = await self._execute_request("api/v1/users/self/")
        return str(check_status_and_parse_json(response)["id"])

    async def get_announcements(self, course_id: str) -> list[DiscussionTopicHeader]:
        self._logger.debug("Fetching announcements for %s", course_id)
        response = await self._execute_request(
            f"api/v1/courses/{course_id}/discussion_topics?only_announcements=true&per_page=999999999999"
        )
        return self._convert_announcements(course_id, response)

    @staticmethod
    def _convert_announcements(
        course_id: str, response: httpx.Response
    ) -> list[DiscussionTopicHeader]:
        # course_id is added to the items for convenience
        return [
            DiscussionTopicHeader(**item, course_id=course_id)
            for item in check_status_and_parse_json(response)
        ]

    async def get_page(self, page_id: str | int, course_id: str | int) -> LegacyPage:
        self._logger.debug(
            "Fetching page content for course %s / page %s", course_id, page_id
        )
        response = await self._execute_request(
            f"api/v1/courses/{course_id}/pages/{page_id}"
        )
        return LegacyPage(**check_status_and_parse_json(response))

    async def get_file(self, file_id: str | int, course_id: str | int) -> RemoteFile:
        self._logger.debug("Fetching file for course %s / file %s", course_id, file_id)
        response = await self._execute_request(
            f"api/v1/courses/{course_id}/files/{file_id}"
        )
        return RemoteFile(**check_status_and_parse_json(response))

    async def get_file_download_stream(self, url: str | URL) -> Response:
        """
        Gets a download stream for a file from canvas. The response is a stream and must be closed with `close()` or `aclose()`
        :param url: The url of the file
        :returns: The response as a stream
        """

        return await self._execute_request(
            endpoint_path=url, stream=True, follow_redirects=True
        )

    async def get_media_object(self, id: str) -> MediaObject:
        self._logger.debug("Fetching media object %s", id)
        response = await self._execute_request(f"media_objects/{id}/info")
        return MediaObject(**check_status_and_parse_json(response))

    async def authenticate_panopto(self, authentication_url: str | URL) -> Response:
        request = self._client.build_request(
            method="GET",
            url=authentication_url,
        )

        return await self._send_with_ratelimit_retry(request, follow_redirects=True)

    async def make_generic_request(
        self, request: Request, follow_redirects: Optional[bool] = None
    ) -> Response:
        """
        This function is only here because of panopto's horrid LTI crap. Avoid using it for anything else.
        """

        self._logger.debug("Making generic request %s", request)

        if await self._is_request_for_canvas(request):
            request.headers.update(self._client_config.get_authorization_header())

        return await self._send_with_ratelimit_retry(
            request, follow_redirects=follow_redirects
        )

    async def _is_request_for_canvas(self, request: httpx.Request) -> bool:
        return request.url.host == self._client_config.host

    async def _execute_request(
        self,
        endpoint_path: str | URL,
        method: str = "GET",
        stream: bool = False,
        follow_redirects: bool = False,
    ) -> Response:
        request = self._client.build_request(
            method=method,
            url=self._client_config.get_endpoint(endpoint_path),
            headers=self._client_config.get_authorization_header(),
        )

        return await self._send_with_ratelimit_retry(
            request=request, stream=stream, follow_redirects=follow_redirects
        )

    @retry(
        wait=wait_exponential_jitter(initial=0.5, exp_base=1.2, max=10),
        retry=retry_if_exception_type(RatelimitedError),
        stop=stop_after_attempt(8),
    )
    async def _send_with_ratelimit_retry(
        self,
        request: Request,
        follow_redirects: Optional[bool] = None,
        stream: bool = False,
    ) -> Response:
        response = await self._authentication_controller.send_and_authenticate(
            request=request, follow_redirects=follow_redirects, stream=stream
        )

        _detect_ratelimit_and_raise(response)

        return response

    @property
    def cookies(self) -> Cookies:
        return self._client.cookies

    async def aclose(self) -> None:
        await self._client.aclose()

    @staticmethod
    async def verify_config(config: CanvasClientConfig) -> None:
        """

        :param config: The config to verify
        :raise ConfigInvalidError: If the url or api key is invalid
        """

        try:
            async with httpx.AsyncClient(http2=True) as client:
                # api/v1/accounts should return very little data, so it's a quick way to verify we are authenticated
                response = await client.get(
                    url=config.get_endpoint("api/v1/accounts"),
                    headers=config.get_authorization_header(),
                    timeout=10,
                )

            response.raise_for_status()
        except httpx.ConnectError as e:
            raise ConfigInvalidError("URL is invalid.") from e
        except httpx.HTTPStatusError as e:
            raise ConfigInvalidError("API token is invalid.") from e


class _CanvasTransport(ExternalClientHTTPXTransport):
    @override
    def _prepare_result(self, response: httpx.Response) -> ExecutionResult:
        _detect_ratelimit_and_raise(response)

        return super()._prepare_result(response)


def _detect_ratelimit_and_raise(response: Response) -> None:
    # Who the FUCK decided to use 403 instead of 429?? With this stupid message??
    # And the newline at the end for some fucking reason is the cherry on top...
    if (
        response.status_code == 403
        and response.text == "403 Forbidden (Rate Limit Exceeded)\n"
        or response.status_code == 429
    ):
        raise RatelimitedError()



from httpx import URL

from libqcanvas_clients.util.url_converter import ensure_is_url


class CanvasClientConfig:
    def __init__(self, api_token: str, canvas_url: str | URL):
        self._api_token = api_token
        self._canvas_url = ensure_is_url(canvas_url)

    def get_authorization_header(self) -> dict[str, str]:
        return {"Authorization": f"Bearer {self._api_token}"}

    def get_endpoint(self, path: URL | str) -> URL:
        return self._canvas_url.join(path)

    @property
    def host(self) -> str:
        return self._canvas_url.host



from httpx import Response

from libqcanvas_clients.util.authenticator import AUTHENTICATION_REQUIRED_RESPONSE_CODE


class CanvasUnauthenticatedDetectorHook:
    def __init__(self, panopto_host: str):
        self.canvas_host = panopto_host

    async def __call__(self, *args, **kwargs):
        response: Response = args[0]

        assert isinstance(response, Response)

        if response.url.host != self.canvas_host:
            return

        # Canvas will silently redirect to the login page or give a 401 if we are not authenticated
        if response.url.path == "/login/canvas":  # or response.status_code == 401
            response.status_code = AUTHENTICATION_REQUIRED_RESPONSE_CODE



# generated by datamodel-codegen:
#   filename:  crap_panopto.json
#   timestamp: 2024-04-28T03:41:49+00:00

from __future__ import annotations

from typing import List, Optional

from pydantic import BaseModel, Field


class PodcastStream(BaseModel):
    absolute_end: float = Field(..., alias="AbsoluteEnd")
    absolute_start: float = Field(..., alias="AbsoluteStart")
    broadcast_type: int = Field(..., alias="BroadcastType")
    edit_media_file_type: None = Field(..., alias="EditMediaFileType")
    edit_media_file_type_legacy: None = Field(..., alias="EditMediaFileTypeLegacy")
    edit_media_file_type_name: None = Field(..., alias="EditMediaFileTypeName")
    interrupted: bool = Field(..., alias="Interrupted")
    name: None = Field(..., alias="Name")
    public_id: str = Field(..., alias="PublicID")
    relative_end: float = Field(..., alias="RelativeEnd")
    relative_segments: None = Field(..., alias="RelativeSegments")
    relative_start: int = Field(..., alias="RelativeStart")
    source_media_file_type: int = Field(..., alias="SourceMediaFileType")
    stream_file_id: str = Field(..., alias="StreamFileId")
    stream_http_url: None = Field(..., alias="StreamHttpUrl")
    stream_type: int = Field(..., alias="StreamType")
    stream_type_name: str = Field(..., alias="StreamTypeName")
    stream_url: str = Field(..., alias="StreamUrl")
    tag: str = Field(..., alias="Tag")
    vr_type: int = Field(..., alias="VRType")
    variants: None = Field(..., alias="Variants")
    viewer_media_file_type: int = Field(..., alias="ViewerMediaFileType")
    viewer_media_file_type_name: str = Field(..., alias="ViewerMediaFileTypeName")


class Variant(BaseModel):
    bandwidth: int = Field(..., alias="Bandwidth")
    url: str = Field(..., alias="Url")


class Stream(BaseModel):
    absolute_end: float = Field(..., alias="AbsoluteEnd")
    absolute_start: float = Field(..., alias="AbsoluteStart")
    broadcast_type: int = Field(..., alias="BroadcastType")
    edit_media_file_type: int = Field(..., alias="EditMediaFileType")
    edit_media_file_type_legacy: None = Field(..., alias="EditMediaFileTypeLegacy")
    edit_media_file_type_name: str = Field(..., alias="EditMediaFileTypeName")
    interrupted: bool = Field(..., alias="Interrupted")
    name: None = Field(..., alias="Name")
    public_id: str = Field(..., alias="PublicID")
    relative_end: float = Field(..., alias="RelativeEnd")
    relative_segments: None = Field(..., alias="RelativeSegments")
    relative_start: int = Field(..., alias="RelativeStart")
    source_media_file_type: int = Field(..., alias="SourceMediaFileType")
    stream_file_id: str = Field(..., alias="StreamFileId")
    stream_http_url: None = Field(..., alias="StreamHttpUrl")
    stream_type: int = Field(..., alias="StreamType")
    stream_type_name: str = Field(..., alias="StreamTypeName")
    stream_url: str = Field(..., alias="StreamUrl")
    tag: str = Field(..., alias="Tag")
    vr_type: int = Field(..., alias="VRType")
    variants: List[Variant] = Field(..., alias="Variants")
    viewer_media_file_type: int = Field(..., alias="ViewerMediaFileType")
    viewer_media_file_type_name: str = Field(..., alias="ViewerMediaFileTypeName")


class Delivery(BaseModel):
    broadcast_ended: bool = Field(..., alias="BroadcastEnded")
    broadcast_interrupted: bool = Field(..., alias="BroadcastInterrupted")
    duration: float = Field(..., alias="Duration")
    has_captions: bool = Field(..., alias="HasCaptions")
    is_active_broadcast: bool = Field(..., alias="IsActiveBroadcast")
    owner_display_name: str = Field(..., alias="OwnerDisplayName")
    owner_id: str = Field(..., alias="OwnerId")
    session_abstract: Optional[str] = Field(..., alias="SessionAbstract")
    session_group_long_name: str = Field(..., alias="SessionGroupLongName")
    session_name: str = Field(..., alias="SessionName")
    session_start_time: float = Field(..., alias="SessionStartTime")
    streams: List[Stream] = Field(..., alias="Streams")
    podcast_streams: List[PodcastStream] = Field(..., alias="PodcastStreams")


class DeliveryInfo(BaseModel):
    completion_percentage: int = Field(..., alias="CompletionPercentage")
    delivery: Delivery = Field(..., alias="Delivery")
    last_viewing_position: int = Field(..., alias="LastViewingPosition")
    session_id: str = Field(..., alias="SessionId")
    session_role: int = Field(..., alias="SessionRole")


class Session(BaseModel):
    field__type: str = Field(..., alias="__type")
    abstract: Optional[str] = Field(..., alias="Abstract")
    analytics_all_time_view_count: int = Field(..., alias="AnalyticsAllTimeViewCount")
    availability_message: str = Field(..., alias="AvailabilityMessage")
    average_rating: int = Field(..., alias="AverageRating")
    bookmarks: List = Field(..., alias="Bookmarks")
    context: List = Field(..., alias="Context")
    contexts_have_time: bool = Field(..., alias="ContextsHaveTime")
    database_session_status: int = Field(..., alias="DatabaseSessionStatus")
    deleted_time: str = Field(..., alias="DeletedTime")
    delivery_id: str = Field(..., alias="DeliveryID")
    delivery_name: str = Field(..., alias="DeliveryName")
    duration: float = Field(..., alias="Duration")
    embed_url: str = Field(..., alias="EmbedUrl")
    folder_id: str = Field(..., alias="FolderID")
    folder_name: str = Field(..., alias="FolderName")
    has_captions: bool = Field(..., alias="HasCaptions")
    has_streams: bool = Field(..., alias="HasStreams")
    has_write_access: bool = Field(..., alias="HasWriteAccess")
    ios_video_url: str = Field(..., alias="IosVideoUrl")
    is_archived: bool = Field(..., alias="IsArchived")
    is_broadcast: bool = Field(..., alias="IsBroadcast")
    is_creator: bool = Field(..., alias="IsCreator")
    is_editable: bool = Field(..., alias="IsEditable")
    is_permanently_deletable: bool = Field(..., alias="IsPermanentlyDeletable")
    is_restorable: bool = Field(..., alias="IsRestorable")
    most_recent_view_position: float = Field(..., alias="MostRecentViewPosition")
    most_recent_viewing: Optional[str] = Field(..., alias="MostRecentViewing")
    notes_url: str = Field(..., alias="NotesUrl")
    owner_full_name: Optional[str] = Field(..., alias="OwnerFullName")
    owner_user_id: Optional[str] = Field(..., alias="OwnerUserId")
    permissions: List[bool] = Field(..., alias="Permissions")
    playable_object_type: int = Field(..., alias="PlayableObjectType")
    podcast_details: Optional[int] = Field(..., alias="PodcastDetails")
    presenter_first_names: List = Field(..., alias="PresenterFirstNames")
    presenter_last_names: List = Field(..., alias="PresenterLastNames")
    presenter_user_names: List = Field(..., alias="PresenterUserNames")
    progress: int = Field(..., alias="Progress")
    remote_recorder_id: Optional[str] = Field(..., alias="RemoteRecorderID")
    remote_recorder_name: Optional[str] = Field(..., alias="RemoteRecorderName")
    session_group_requires_approval: bool = Field(
        ..., alias="SessionGroupRequiresApproval"
    )
    session_id: str = Field(..., alias="SessionID")
    session_name: str = Field(..., alias="SessionName")
    show_settings: bool = Field(..., alias="ShowSettings")
    start_time: str = Field(..., alias="StartTime")
    status: int = Field(..., alias="Status")
    tablet_video_url: str = Field(..., alias="TabletVideoUrl")
    tags: List = Field(..., alias="Tags")
    thumb_url: str = Field(..., alias="ThumbUrl")
    user_can_enumerate_folder: bool = Field(..., alias="UserCanEnumerateFolder")
    viewer_url: str = Field(..., alias="ViewerUrl")
    editor_url: str = Field(..., alias="EditorUrl")
    order_sort: int = Field(..., alias="OrderSort")


class SessionList(BaseModel):
    field__type: str = Field(..., alias="__type")
    results: List[Session] = Field(..., alias="Results")
    total_number: int = Field(..., alias="TotalNumber")
    archived_count: Optional[int] = Field(..., alias="ArchivedCount")
    can_add_subfolder: bool = Field(..., alias="CanAddSubfolder")
    contains_archived_deliveries: bool = Field(..., alias="ContainsArchivedDeliveries")
    folder_type: int = Field(..., alias="FolderType")


class Parent(BaseModel):
    is_editable: bool = Field(..., alias="IsEditable")
    is_department: bool = Field(..., alias="IsDepartment")
    is_dropbox: bool = Field(..., alias="IsDropbox")
    folder_type: int = Field(..., alias="FolderType")
    deliveries_have_specified_order: bool = Field(
        ..., alias="DeliveriesHaveSpecifiedOrder"
    )
    podcast_enabled: bool = Field(..., alias="PodcastEnabled")
    user_can_create_folders: bool = Field(..., alias="UserCanCreateFolders")
    user_can_create_sessions: bool = Field(..., alias="UserCanCreateSessions")
    user_can_manage_department_settings: bool = Field(
        ..., alias="UserCanManageDepartmentSettings"
    )
    user_can_move_folder: bool = Field(..., alias="UserCanMoveFolder")
    user_can_move_session_destination: bool = Field(
        ..., alias="UserCanMoveSessionDestination"
    )
    is_enumerable: bool = Field(..., alias="IsEnumerable")
    has_accessible_children: bool = Field(..., alias="HasAccessibleChildren")
    repository_key: str = Field(..., alias="RepositoryKey")
    id: str = Field(..., alias="Id")
    name: Optional[str] = Field(..., alias="Name")
    type_name: str = Field(..., alias="TypeName")


class FolderInfo(BaseModel):
    is_editable: bool = Field(..., alias="IsEditable")
    is_department: bool = Field(..., alias="IsDepartment")
    is_dropbox: bool = Field(..., alias="IsDropbox")
    folder_type: int = Field(..., alias="FolderType")
    deliveries_have_specified_order: bool = Field(
        ..., alias="DeliveriesHaveSpecifiedOrder"
    )
    podcast_enabled: bool = Field(..., alias="PodcastEnabled")
    parent: Optional[Parent] = Field(..., alias="Parent")
    user_can_create_folders: bool = Field(..., alias="UserCanCreateFolders")
    user_can_create_sessions: bool = Field(..., alias="UserCanCreateSessions")
    user_can_manage_department_settings: bool = Field(
        ..., alias="UserCanManageDepartmentSettings"
    )
    user_can_move_folder: bool = Field(..., alias="UserCanMoveFolder")
    user_can_move_session_destination: bool = Field(
        ..., alias="UserCanMoveSessionDestination"
    )
    is_enumerable: bool = Field(..., alias="IsEnumerable")
    has_accessible_children: bool = Field(..., alias="HasAccessibleChildren")
    session_count: int = Field(..., alias="SessionCount")
    repository_key: str = Field(..., alias="RepositoryKey")
    id: str = Field(..., alias="Id")
    name: str = Field(..., alias="Name")
    type_name: str = Field(..., alias="TypeName")


class AlternativeParent(BaseModel):
    id: str = Field(..., alias="Id")
    name: str = Field(..., alias="Name")
    type_name: str = Field(..., alias="TypeName")


class AlternativeSession(BaseModel):
    parent: AlternativeParent = Field(..., alias="Parent")
    viewer_uri: str = Field(..., alias="ViewerUri")
    embed_viewer_tag: str = Field(..., alias="EmbedViewerTag")
    is_webcast: bool = Field(..., alias="IsWebcast")
    description: Optional[str] = Field(None, alias="Description")
    status: int = Field(..., alias="Status")
    start_time: str = Field(..., alias="StartTime")
    duration: float = Field(..., alias="Duration")
    thumb_url: str = Field(..., alias="ThumbUrl")
    most_recent_view_time: Optional[str] = Field(None, alias="MostRecentViewTime")
    is_reference_copy: bool = Field(..., alias="IsReferenceCopy")
    repository_key: str = Field(..., alias="RepositoryKey")
    id: str = Field(..., alias="Id")
    name: str = Field(..., alias="Name")
    type_name: str = Field(..., alias="TypeName")



class RatelimitedError(Exception):
    pass


class UnauthenticatedError(Exception):
    pass


class AuthenticationFailedError(Exception):
    def __init__(self, msg: str):
        super().__init__(msg)


class RequestFailedError(Exception):
    def __init__(self, msg: str):
        super().__init__(msg)


class ConfigInvalidError(Exception):
    def __init__(self, msg: str):
        super().__init__(msg)



from httpx import URL

from libqcanvas_clients.util.url_converter import ensure_is_url


class PanoptoClientConfig:
    def __init__(self, panopto_url: str | URL):
        self._panopto_url = ensure_is_url(panopto_url)

    def get_endpoint(self, path: URL | str) -> URL:
        return self._panopto_url.join(path)

    @property
    def host(self) -> str:
        return self._panopto_url.host



import logging

from libqcanvas_clients.canvas.canvas_client import CanvasClient
from libqcanvas_clients.panopto.panopto_client_config import PanoptoClientConfig
from libqcanvas_clients.util.authenticator import Authenticator
from libqcanvas_clients.util.request_exceptions import AuthenticationFailedError

_logger = logging.getLogger(__name__)


class PanoptoAuthenticator(Authenticator):
    def __init__(
        self, panopto_client_config: PanoptoClientConfig, canvas_client: CanvasClient
    ):
        super().__init__(canvas_client._client)
        self.max_retries = 5
        self._panopto_client_config = panopto_client_config
        self._canvas_client = canvas_client

    async def _authenticate(self) -> None:
        _logger.info("Authenticating to panopto")

        response = await self._canvas_client.authenticate_panopto(
            self._panopto_client_config.get_endpoint(
                "Panopto/Pages/Auth/Login.aspx?instance=Canvas&AllowBounce=true"
            )
        )

        response.raise_for_status()
        _logger.debug("Got response from %s", response.url)

        if not response.url.path.endswith("Home.aspx"):
            _logger.error("Panopto authentication failed")
            raise AuthenticationFailedError(
                "Couldn't authenticate to panopto via canvas. Is your canvas account linked to panopto?"
            )
        else:
            _logger.info("Authentication complete")



import logging

from httpx import Response

from libqcanvas_clients.util.authenticator import AUTHENTICATION_REQUIRED_RESPONSE_CODE

_logger = logging.getLogger(__name__)


class PanoptoUnauthenticatedDetectorHook:
    def __init__(self, panopto_host: str):
        self.panopto_host = panopto_host

    async def __call__(self, *args, **kwargs) -> None:
        response: Response = args[0]

        assert isinstance(response, Response)

        if response.url.host != self.panopto_host:
            return

        auth_cookie = ".ASPXAUTH"
        if not response.is_server_error and not response.is_redirect:
            no_auth_cookie = auth_cookie not in response.cookies.keys()
            if no_auth_cookie or response.cookies.get(auth_cookie) == "":
                if no_auth_cookie:
                    reason = "auth cookie missing"
                else:
                    reason = "auth cookie empty"

                _logger.info("Panopto request was unauthorised because %s", reason)
                response.status_code = AUTHENTICATION_REQUIRED_RESPONSE_CODE



from .canvas_client import CanvasClient
from .canvas_client_config import CanvasClientConfig
from .canvas_types import DiscussionTopicHeader as Announcement
from .canvas_types import LegacyPage, RemoteFile
